#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import re
import sys
import copy
import json
import flask
import string
import locale
import random
import datetime
import itertools
import threading
import subprocess

import jinja2

from . import legacy
from . import common
from . import defines
from . import exceptions

FIRST_CAP_REGEX = re.compile("(.)([A-Z][a-z]+)")
""" Regular expression that ensures that the first
token of each camel string is properly capitalized """

ALL_CAP_REGEX = re.compile("([a-z0-9])([A-Z])")
""" The generalized transition from lower case to
upper case letter regex that will provide a way of
putting the underscore in the middle of the transition """

SORT_MAP = {
    "1" : 1,
    "-1" : -1,
    "ascending" : 1,
    "descending" : -1,
}
""" The map associating the normalized (text) way of
representing sorting with the current infra-structure
number way of representing the same information """

CASTERS = {
    list : lambda v: [y for y in itertools.chain(*[x.split(",") for x in v])],
    bool : lambda v: v if isinstance(v, bool) else\
        not v in ("", "0", "false", "False")
}
""" The map associating the various data types with a proper custom
caster to be used for special data types (more complex) under some
of the simple casting operations """

defines = defines

def to_limit(limit_s):
    limit = int(limit_s)
    if limit < 0: return 0
    return limit

def to_find(find_s):
    if not find_s: return []
    find_t = type(find_s)
    if find_t == list: return find_s
    return [find_s]

def to_sort(sort_s):
    values = sort_s.split(":", 1)
    if len(values) == 1: values.append("descending")
    name, direction = values
    if name == "default": return None
    values[1] = SORT_MAP.get(direction, 1)
    return [tuple(values)]

ALIAS = {
    "context" : "find_d",
    "filters" : "find_d",
    "filters[]" : "find_d",
    "filter_def" : "find_d",
    "filter_string" : "find_s",
    "filter_name" : "find_n",
    "insensitive" : "find_i",
    "order" : "sort",
    "offset" : "skip",
    "start_record" : "skip",
    "number_records" : "limit"
}
""" The map containing the various attribute alias
between the normalized manned and the quorum manner """

FIND_TYPES = dict(
    skip = int,
    limit = to_limit,
    find_s = legacy.UNICODE,
    find_d = to_find,
    find_i = bool,
    find_t = legacy.UNICODE,
    find_n = legacy.UNICODE,
    sort = to_sort,
    meta = bool,
    fields = list
)
""" The map associating the various find fields with
their respective types, note that in case a special
conversion operation is required the associated value
may represent a conversion function instead """

FIND_DEFAULTS = dict(
    limit = 10
)
""" The map that defines the various default values
for a series of find related attributes """

def is_iterable(object):
    return isinstance(object, defines.ITERABLES)

def request_json(request = None, encoding = "utf-8"):
    # retrieves the proper request object, either the provided
    # request or the default flask request object and then in
    # case the the json data is already in the request properties
    # it is used (cached value) otherwise continues with the parse
    request = request or flask.request
    try:
        if "_data_j" in request.properties:
            return request.properties["_data_j"]
    except RuntimeError: pass

    # retrieves the current request data and tries to
    # "load" it as json data, in case it fails gracefully
    # handles the failure setting the value as an empty map
    data = request.data
    try:
        is_bytes = legacy.is_bytes(data)
        if is_bytes: data = data.decode(encoding)
        data_j = json.loads(data)
    except: data_j = {}
    request.properties["_data_j"] = data_j

    # returns the json data object to the caller method so that it
    # may be used as the parsed value (post information)
    return data_j

def get_field(name, default = None, cast = None, strip = False):
    # tries to retrieve the json based representation of the provided
    # request from all the possible sources, this is required because
    # it's going to be used to try to retrieve a field from it
    data_j = request_json()

    # tries to retrieve the requested field from all the requested sources
    # the order of retrieval should respect the importance of each of the
    # sources (from the least important to the most important)
    value = data_j.get(name, default)
    value = flask.request.files.get(name, value)
    value = flask.request.form.get(name, value)
    value = flask.request.args.get(name, value)

    # in case the strip flag is enabled an "extra" strip operation is applied
    # to the retrieved value (as requested), avoiding extra spaces
    if strip: value = value.strip()

    # in case a cast operation is defined, tries to retrieve a possible
    # indirect/custom caster for the current cast operation
    if cast: cast = CASTERS.get(cast, cast)

    # in case the cast type value is set and the value is not invalid tries
    # to cast the value into the requested cast type
    if cast and not value in (None, ""): value = cast(value)

    # returns the final value to the caller method to be used, note that the
    # caller method should be aware of the sources used in the field retrieval
    return value

def get_object(
    object = None,
    alias = False,
    page = False,
    find = False,
    norm = True,
    **kwargs
):
    # verifies if the provided object is valid in such case creates
    # a copy of it and uses it as the base object for validation
    # otherwise used an empty map (form validation)
    object = object and copy.copy(object) or {}

    # retrieves the current request data and tries to
    # "load" it as json data, in case it fails gracefully
    # handles the failure setting the value as an empty map
    try: data_j = request_json()
    except RuntimeError: data_j = dict()

    # retrieves the reference to the file objects currently
    # present in the request, note that in case a runtime
    # error occurs (eg: out of context) fails gracefully
    try: files = flask.request.files
    except RuntimeError: files = dict()

    # retrieves the reference to the form objects currently
    # present in the request, note that in case a runtime
    # error occurs (eg: out of context) fails gracefully
    try: form_s = flask.request.form_s
    except RuntimeError: form_s = dict()

    # retrieves the reference to the arguments objects currently
    # present in the request, note that in case a runtime
    # error occurs (eg: out of context) fails gracefully
    try: args_s = flask.request.args_s
    except RuntimeError: args_s = dict()

    # uses all the values referencing data in the request to try
    # to populate the object this way it may be constructed using
    # any of theses strategies (easier for the developer)
    for name, value in data_j.items(): object[name] = value
    for name, value in files.items(): object[name] = value
    for name, value in form_s.items(): object[name] = value
    for name, value in args_s.items(): object[name] = value

    # in case the alias flag is set tries to resolve the attribute
    # alias and in case the find types are set converts the find
    # based attributes using the currently defined mapping map
    alias and resolve_alias(object)
    page and page_types(object)
    find and find_types(object)
    find and find_defaults(object, kwargs)

    # in case the normalization flag is set runs the normalization
    # of the provided object so that sequences are properly handled
    # as defined in the specification (this allows multiple references)
    norm and norm_object(object)

    # returns the constructed object to the caller method this object
    # should be a structured representation of the data in the request
    return object

def is_mobile(user_agent = None):
    """
    Verifies if the current user agent string represents a
    mobile agent, for that a series of regular expressions
    are matched against the user agent string.

    :type user_agent: String
    :param user_agent: The string containing the user agent
    value that is going to be verified against a series of
    regular expressions for mobile verification.
    :rtype: bool
    :return: If the current user agent string represents a
    mobile browser or a regular (desktop) one.
    """

    user_agent = flask.request.headers.get("User-Agent", "")\
        if user_agent == None else user_agent
    prefix = user_agent[:4]
    mobile = defines.MOBILE_REGEX.search(user_agent)
    mobile_prefix = defines.MOBILE_PREFIX_REGEX.search(prefix)
    is_mobile = True if mobile or mobile_prefix else False
    return is_mobile

def is_tablet(user_agent = None):
    """
    Verifies if the current user agent string represents a
    tablet agent, for that a series of regular expressions
    are matched against the user agent string.

    :type user_agent: String
    :param user_agent: The string containing the user agent
    value that is going to be verified against a series of
    regular expressions for tablet verification.
    :rtype: bool
    :return: If the current user agent string represents a
    tablet browser or a regular (desktop) one.
    """

    user_agent = flask.request.headers.get("User-Agent", "")\
        if user_agent == None else user_agent
    prefix = user_agent[:4]
    tablet = defines.TABLET_REGEX.search(user_agent)
    mobile_prefix = defines.MOBILE_PREFIX_REGEX.search(prefix)
    is_tablet = True if tablet or mobile_prefix else False
    return is_tablet

def is_browser(user_agent = None):
    """
    Verifies if the provided user agent string represents a
    browser (interactive) agent, for that a series of verifications
    are going to be performed against the user agent string.

    :type user_agent: String
    :param user_agent: The string containing the user agent
    value that is going to be verified for browser presence.
    :rtype: bool
    :return: If the provided user agent string represents an
    interactive browser or not.
    """

    user_agent = flask.request.headers.get("User-Agent", "")\
        if user_agent == None else user_agent
    info = browser_info(user_agent = user_agent)
    if not info: return False
    interactive = info.get("interactive", False)
    if not interactive: return False
    return True

def is_bot(user_agent = None):
    """
    Verifies if the provided user agent string represents a
    bot (automated) agent, for that a series of verifications
    are going to be performed against the user agent string.

    :type user_agent: String
    :param user_agent: The string containing the user agent
    value that is going to be verified for bot presence.
    :rtype: bool
    :return: If the provided user agent string represents an
    automated bot or not.
    """

    user_agent = flask.request.headers.get("User-Agent", "")\
        if user_agent == None else user_agent
    info = browser_info(user_agent = user_agent)
    if not info: return False
    bot = info.get("bot", False)
    if not bot: return False
    return True

def browser_info(user_agent = None):
    """
    Retrieves a dictionary containing information about the browser
    and the operative system associated with the provided user agent.

    The retrieval of the information depends on the kind of user
    agent string provided, as coverage is limited.

    :type user_agent: String
    :param user_agent: The HTTP based user agent string to be processed.
    :rtype: Dictionary
    :return: The dictionary/map containing the information processed from
    the provided user agent.
    """

    user_agent = flask.request.headers.get("User-Agent", "")\
        if user_agent == None else user_agent

    info = dict()

    for browser_i in defines.BROWSER_INFO:
        identity = browser_i["identity"]
        sub_string = browser_i.get("sub_string", identity)
        version_search = browser_i.get("version_search", sub_string + "/")
        interactive = browser_i.get("interactive", True)
        bot = browser_i.get("bot", False)

        if not sub_string in user_agent: continue
        if not version_search in user_agent: continue

        version_i = user_agent.index(version_search) + len(version_search)
        version = user_agent[version_i:].split(" ", 1)[0].strip(" ;")
        version_f = float(".".join(version.split(".")[:2]))
        version_i = int(version_f)

        info.update(
            name = identity,
            version = version,
            version_f = version_f,
            version_i = version_i,
            interactive = interactive,
            bot = bot
        )
        break

    for os_i in defines.OS_INFO:
        identity = os_i["identity"]
        sub_string = os_i.get("sub_string", identity)

        if not sub_string in user_agent: continue

        info.update(os = identity)
        break

    return info if info else None

def resolve_alias(object):
    for name, value in legacy.eager(object.items()):
        if not name in ALIAS: continue
        _alias = ALIAS[name]
        object[_alias] = value
        del object[name]

def page_types(object, size = 50):
    page = object.get("page", 1)
    size = object.get("size", size)
    sorter = object.get("sorter", None)
    direction = object.get("direction", "descending")
    page = int(page)
    size = int(size)
    offset = page - 1
    object["skip"] = offset * size
    object["limit"] = size
    if sorter: object["sort"] = "%s:%s" % (sorter, direction)

def find_types(object):
    # iterates over all the name and values of the object
    # trying to convert each of the find types into the
    # appropriate representation for the infra-structure
    for name, value in legacy.eager(object.items()):
        # in case the current find name is not present
        # in the find types map it's removed from the
        # object as it is considered to be invalid
        if not name in FIND_TYPES:
            del object[name]
            continue

        # retrieves the find (data) type associated with
        # the current find name and "runs" the conversion
        # method retrieving the resulting value
        find_type = FIND_TYPES[name]
        value = find_type(value)

        # in case the value resulting from the conversion
        # is invalid the name is removed from the find object
        # because it's not considered to be valid
        if value == None: del object[name]
        else: object[name] = value

def find_defaults(object, kwargs):
    for name, value in legacy.iteritems(kwargs):
        if name in object: continue
        if not name in FIND_TYPES: continue
        object[name] = value

    for name, value in legacy.iteritems(FIND_DEFAULTS):
        if name in object: continue
        object[name] = value

def norm_object(object):
    # iterates over all the key value association in the
    # object, trying to find the ones that refer sequences
    # so that they may be normalized
    for name, value in legacy.eager(object.items()):
        # verifies if the current name references a sequence
        # and if that's not the case continues the loop trying
        # to find any other sequence based value
        if not name.endswith("[]"): continue

        # removes the current reference to the name as the value
        # is not in the valid structure and then normalizes the
        # name by removing the extra sequence indication value
        del object[name]
        name = name[:-2]

        # in case the current value is not valid (empty) the object
        # is set with an empty list for the current iteration as this
        # is considered to be the default value
        if not value: object[name] = []; continue

        # retrieves the normalized and linearized list of leafs
        # for the current value and ten verifies the size of each
        # of its values and uses it to measure the number of
        # dictionary elements that are going to be contained in
        # the sequence to be "generated", then uses this (size)
        # value to pre-generate the complete set of dictionaries
        leafs_l = leafs(value)
        first = leafs_l[0] if leafs_l else (None, [])
        _fqn, values = first
        size = len(values)
        list = [dict() for _index in range(size)]

        # sets the list of generates dictionaries in the object for
        # the newly normalized name of structure
        object[name] = list

        # iterates over the complete set of key value pairs in the
        # leafs list to gather the value into the various objects that
        # are contained in the sequence (normalization process)
        for _name, _value in leafs_l:
            for index in range(size):
                _object = list[index]
                _name_l = _name.split(".")
                set_object(_object, _name_l, _value[index])

def set_object(object, name_l, value):
    """
    Sets a composite value in an object, allowing for
    dynamic setting of random size key values.

    This method is useful for situations where one wants
    to set a value at a randomly defined depth inside
    an object without having to much work with the creation
    of the inner dictionaries.

    :type object: Dictionary
    :param object: The target object that is going to be
    changed and set with the target value.
    :type name_l: List
    :param name_l: The list of names that defined the fully
    qualified name to be used in the setting of the value
    for example path.to.end will be a three size list containing
    each of the partial names.
    :type value: Object
    :param value: The value that is going to be set in the
    defined target of the object.
    """

    # retrieves the first name in the names list this is the
    # value that is going to be used for the current iteration
    name = name_l[0]

    # in case the length of the current names list has reached
    # one this is the final iteration and so the value is set
    # at the current naming point
    if len(name_l) == 1: object[name] = value

    # otherwise this is a "normal" step and so a new map must
    # be created/retrieved and the iteration step should be
    # performed on this new map as it's set on the current naming
    # place (recursion step)
    else:
        map = object.get(name, {})
        object[name] = map
        set_object(map, name_l[1:], value)

def leafs(object):
    """
    Retrieves a list containing a series of tuples that
    each represent a leaf of the current object structure.

    A leaf is the last element of an object that is not a
    map, the other intermediary maps are considered to be
    trunks and should be percolated recursively.

    This is a recursive function that takes some memory for
    the construction of the list, and so should be used with
    the proper care to avoid bottlenecks.

    :type object: Dictionary
    :param object: The object for which the leafs list
    structure is meant to be retrieved.
    :rtype: List
    :return: The list of leaf node tuples for the provided
    object, as requested for each of the sequences.
    """

    # creates the list that will hold the various leaf nodes
    # "gathered" by the current recursion function
    leafs_l = []

    # iterates over all the key and value relations in the
    # object trying to find the leaf nodes (no map nodes)
    # creating a tuple of fqn (fully qualified name) and value
    for name, value in object.items():
        # retrieves the data type for the current value and
        # validation if it is a dictionary or any other type
        # in case it's a dictionary a new iteration step must
        # be performed retrieving the leafs of the value and
        # then incrementing the name with the current prefix
        value_t = type(value)
        if value_t == dict:
            _leafs = leafs(value)
            _leafs = [(name + "." + _name, value) for _name, value in _leafs]
            leafs_l.extend(_leafs)

        # otherwise this is a leaf node and so the leaf tuple
        # node must be constructed with the current value
        # (properly validated for sequence presence)
        else:
            value_t = type(value)
            if not value_t == list: value = [value]
            leafs_l.append((name, value))

    # returns the list of leaf nodes that was "just" created
    # to the caller method so that it may be used there
    return leafs_l

def load_form(form):
    # creates the map that is going to hold the "structured"
    # version of the form with key value associations
    form_s = dict()

    # iterates over all the form items to parse their values
    # and populate the form structured version of it, note that
    # for the sake of parsing the order of the elements in the
    # form is relevant, in case there's multiple values for the
    # same name they are considered as a list, otherwise they are
    # considered as a single value
    for name in form:
        # retrieves the value (as a list) for the current name, then
        # in case the sequence is larger than one element sets it,
        # otherwise retrieves and sets the value as the first element
        value = form.getlist(name)
        value = value[0] if isinstance(value, (list, tuple)) and\
            len(value) == 1 else value

        # splits the complete name into its various components
        # and retrieves both the final (last) element and the
        # various partial elements from it
        names = name.split(".")
        final = names[-1]
        partials = names[:-1]

        # sets the initial "struct" reference as the form structured
        # that has just been created (initial structure for iteration)
        # then starts the iteration to retrieve or create the various
        # intermediate structures
        struct = form_s
        for _name in partials:
            _struct = struct.get(_name, {})
            struct[_name] = _struct
            struct = _struct

        # sets the current value in the currently loaded "struct" element
        # so that the reference gets properly updated
        struct[final] = value

    # retrieves the final "normalized" form structure containing
    # a series of chained maps resulting from the parsing of the
    # linear version of the attribute names
    return form_s

def load_locale(available, fallback = "en_us"):
    # tries to gather the best locale value using the currently
    # available strategies and in case the retrieved local is part
    # of the valid locales for the app returns the locale, otherwise
    # returns the fallback value instead
    locale = get_locale(fallback = fallback)
    if locale in available: return locale
    return fallback

def get_locale(fallback = "en_us"):
    # tries to retrieve the locale value from the provided URL
    # parameters (this is the highest priority) and in case it
    # exists returns this locale immediately
    locale = flask.request.args.get("locale", None)
    if locale: return locale

    # uses the currently loaded session to try to gather the locale
    # value from it and in case it's valid and exists returns it
    locale = flask.session.get("locale", None)
    if locale: return locale

    # gathers the complete set of language values set in the accept
    # language header and in case there's at least one value returned
    # returns the first of these values as the locale
    langs = get_langs()
    if langs: return langs[0]

    # in case this code entry is reached all the strategies for locale
    # retrieval have failed and so the fallback value is returned
    return fallback

def get_langs():
    # gathers the value of the accept language header and in case
    # it's not defined returns immediately as no language can be
    # determined using the currently provided headers
    accept_language = flask.request.headers.get("Accept-Language", None)
    if not accept_language: return ()

    # starts the list that is going to be used to store the various
    # languages "recovered" from the accept language header, note that
    # the order of these languages should be from the most relevant to
    # the least relevant as defined in http specification
    langs = []

    # splits the accept language header into the various components of
    # it and then iterates over each of them splitting each of the
    # components into the proper language string and priority
    parts = accept_language.split(",")
    for part in parts:
        values = part.split(";", 1)
        value_l = len(values)
        if value_l == 1: lang, = values
        else: lang, _priority = values
        lang = lang.replace("-", "_")
        lang = lang.lower()
        langs.append(lang)

    # returns the complete list of languages that have been extracted
    # from the accept language header these list may be empty in case
    # the header was not parsed correctly or there's no contents in it
    return langs

def set_locale():
    # normalizes the current locale string by converting the
    # last part of the locale string to an uppercase representation
    # and then re-joining the various components of it
    values = flask.request.locale.split("_", 1)
    if len(values) > 1: values[1] = values[1].upper()
    locale_n = "_".join(values)
    locale_n = str(locale_n)

    # in case the current operative system is windows based an
    # extra locale conversion operation must be performed, after
    # than the proper setting of the os locale is done with the
    # fallback for exception being silent (non critical)
    if os.name == "nt": locale_n = defines.WINDOWS_LOCALE.get(locale_n, "")
    else: locale_n += ".utf8"
    try: locale.setlocale(locale.LC_ALL, locale_n)
    except: pass

def reset_locale():
    locale.setlocale(locale.LC_ALL, "")

def anotate_async(response):
    # verifies if the current response contains the location header
    # meaning that a redirection will occur, and if that's not the
    # case this function returns immediately to avoid problems
    if not "Location" in response.headers: return

    # checks if the current request is "marked" as asynchronous, for
    # such cases a special redirection process is applies to avoid the
    # typical problems with automated redirection using "ajax"
    is_async = True if flask.request.headers.get("X-Async") else False
    is_async = True if get_field("async") else is_async
    if is_async: response.status_code = 280

def run_thread(function, *args, **kwargs):
    return threading.Thread(
        target = function,
        args = args,
        kwargs = kwargs
    ).start()

def camel_to_underscore(camel, separator = "_", lower = True):
    """
    Converts the provided camel cased based value into
    a normalized underscore based string.

    An optional lower parameter may be used to avoid the case
    of the letters from being lower cased.

    This is useful as most of the python string standards
    are compliant with the underscore strategy.

    :type camel: String
    :param camel: The camel cased string that is going to be
    converted into an underscore based string.
    :type separator: String
    :param separator: The separator token that is going to
    be used in the camel to underscore conversion.
    :type lower: bool
    :param lower: If the letter casing should be changed while
    convert the value from camel to underscore.
    :rtype: String
    :return: The underscore based string resulting from the
    conversion of the provided camel cased one.
    """

    value = FIRST_CAP_REGEX.sub(r"\1" + separator + r"\2", camel)
    value = ALL_CAP_REGEX.sub(r"\1" + separator + r"\2", value)
    if lower: value = value.lower()
    return value

def camel_to_readable(camel, lower = False, capitalize = False):
    """
    Converts the given camel cased oriented string value
    into a readable one meaning that the returned value
    is a set of strings separated by spaces.

    This method may be used to convert class names into
    something that is readable by an end user.

    :type camel: String
    :param camel: The camel case string value that is going
    to be used in the conversion into a readable string.
    :type lower: bool
    :param lower: If the camel based value should be lower
    cased before the conversion to readable.
    :type capitalize: bool
    :param capitalize: If all of the words should be capitalized
    or if instead only the first one should.
    :rtype: String
    :return: The final human readable string that may be
    used to display a value to an end user.
    """

    underscore = camel_to_underscore(camel, lower = lower)
    return underscore_to_readable(underscore, capitalize = capitalize)

def underscore_to_readable(underscore, capitalize = False):
    """
    Converts the given underscore oriented string value
    into a readable one meaning that the returned value
    is a set of strings separated by spaces.

    This method may be used to class attributes into
    something that is readable by an end user.

    :type camel: String
    :param camel: The underscore string value that is going
    to be used in the conversion into a readable string.
    :type capitalize: bool
    :param capitalize: If all of the words should be capitalized
    or if instead only the first one should.
    :rtype: String
    :return: The final human readable string that may be
    used to display a value to an end user.
    """

    parts = underscore.split("_")
    if capitalize: parts = [part[0].upper() + part[1:] for part in parts]
    else: parts[0] = parts[0][0].upper() + parts[0][1:]
    return " ".join(parts)

def generate_identifier(size = 16, chars = string.ascii_uppercase + string.digits):
    """
    Generates a random identifier (may be used as password) with
    the provided constrains of length and character ranges.

    This function may be used in the generation of random based
    keys for both passwords and api keys.

    :type size: int
    :param size: The size (in number of characters) to be used in
    the generation of the random identifier.
    :type chars: List
    :param chars: The list containing the characters to be used
    in the generation of the random identifier.
    :rtype: String
    :return: The randomly generated identifier obeying to the
    provided constrains.
    """

    return "".join(random.choice(chars) for _index in range(size))

def to_locale(value, locale = None, fallback = True):
    """
    Utility function used to localize the provided value according
    to the currently loaded set of bundles, the bundles are loaded
    at the application start time from the proper sources.

    The (target) locale value for the translation may be provided or
    in case it's not the locale associated with the current request
    is used as an alternative.

    In case the value is not localizable (no valid bundle available)
    it is returned as it is without change.

    :type value: String
    :param value: The value that is going to be localized according
    to the current application environment, this may be a normal
    english dictionary string or a variable reference.
    :type locale: String
    :param locale: The (target) locale value to be used in the
    translation process for the provided string value.
    :type fallback: bool
    :param fallback: If a fallback operation should be performed in
    case no value was retrieved from the base/request locale.
    :rtype: String
    :return: The localized value for the current environment or the
    proper (original) value in case no localization was possible.
    """

    value_t = type(value)
    is_sequence = value_t in (list, tuple)
    if is_sequence: return _serialize([
        to_locale(value, locale = locale, fallback = fallback)\
        for value in value
    ])
    has_context = common.base().has_context()
    locale = locale or (flask.request.locale if has_context else None)
    bundle = common.base().get_bundle(locale) or {}
    result = bundle.get(value, None)
    if not result == None: return result
    app = common.base().APP
    if fallback and app: return to_locale(
        value,
        locale = app._locale_d,
        fallback = False
    )
    return value

def nl_to_br(value):
    """
    Replaces the occurrences of the newline character in the
    string with the HTML break line character.

    This is useful for one trying to convert a plain text
    representation of a string into HTML representation.

    :type value: String
    :param value: The base value that is going to be used in
    the conversion to the HTML value.
    :rtype: String
    :return: The string containing the newline characters replaced
    with line breaking HTML tags.
    """

    return value.replace("\n", "<br/>\n")

def nl_to_br_jinja(eval_ctx, value):
    """
    Optimized version of the function that replaces newline
    characters with the HTML break lines that handles the
    autoescape properties of the jinja engine.

    :type eval_ctx: Context
    :param eval_ctx: Current evaluation context being used
    in the rendering of the jinja template. May be used to
    determine if the autoescape mode is being used.
    :type value: String
    :param value: The base value that is going to be used in
    the conversion to the html value.
    :rtype: String
    :return: The string containing the newline characters replaced
    with line breaking HTML tags.
    """

    if eval_ctx.autoescape: value = legacy.UNICODE(jinja2.escape(value))
    value = nl_to_br(value)
    if eval_ctx.autoescape: value = jinja2.Markup(value)
    return value

def sp_to_nbsp(value):
    """
    Replaces the occurrences of the space character in the
    string with the HTML non breaking space character.

    This is useful for one trying to convert a plain text
    representation of a string into HTML representation.

    :type value: String
    :param value: The base value that is going to be used in
    the conversion to the HTML value.
    :rtype: String
    :return: The string containing the space characters replaced
    with non breaking space characters for HTML.
    """

    return value.replace(" ", "&nbsp;")

def sp_to_nbsp_jinja(eval_ctx, value):
    """
    Optimized version of the function that replaces space
    characters with the HTML non breaking space characters
    that handles the autoescape properties of the jinja engine.

    :type eval_ctx: Context
    :param eval_ctx: Current evaluation context being used
    in the rendering of the jinja template. May be used to
    determine if the autoescape mode is being used.
    :type value: String
    :param value: The base value that is going to be used in
    the conversion to the HTML value.
    :rtype: String
    :return: The string containing the space characters replaced
    with non breaking space characters for HTML.
    """

    if eval_ctx.autoescape: value = legacy.UNICODE(jinja2.escape(value))
    value = sp_to_nbsp(value)
    if eval_ctx.autoescape: value = jinja2.Markup(value)
    return value

def date_time(value, format = "%d/%m/%Y"):
    """
    Formats the provided as a date string according to the
    provided date format.

    Assumes that the provided value represents a float string
    and that may be used as the based timestamp for conversion.

    :type value: String
    :param value: The base timestamp value string that is going
    to be used for the conversion of the date string.
    :type format: String
    :param format: The format string that is going to be used
    when formatting the date time value.
    :rtype: String
    :return: The resulting date time string that may be used
    to represent the provided value.
    """

    # tries to convert the provided string value into a float
    # in case it fails the proper string value is returned
    # immediately as a fallback procedure
    try: value_f = float(value)
    except: return value

    # creates the date time structure from the provided float
    # value and then formats the date time according to the
    # provided format and returns the resulting string
    date_time_s = datetime.datetime.utcfromtimestamp(value_f)
    return date_time_s.strftime(format).decode("utf-8")

def quote(value, *args, **kwargs):
    """
    Quotes the passed value according to the defined
    standard for url escaping, the value is first encoded
    into the expected utf-8 encoding as defined by standard.

    This method should be used instead of a direct call to
    the equivalent call in the URL library.

    :type value: String
    :param value: The string value that is going to be quoted
    according to the URL escaping scheme.
    :rtype: String
    :return: The quoted value according to the URL scheme this
    value may be safely used in urls.
    """

    is_unicode = isinstance(value, legacy.UNICODE)
    if is_unicode: value = value.encode("utf-8")
    return legacy.quote(value, *args, **kwargs)

def unquote(value, *args, **kwargs):
    """
    Unquotes the provided value according to the URL scheme
    the resulting value should be an unicode string representing
    the same value, the intermediary string value from the decoding
    should be an utf-8 based value.

    This method should be used instead of a direct call to
    the equivalent call in the URL library.

    :type value: String
    :param value: The string value that is going to be unquoted
    according to the URL escaping scheme.
    :rtype: String
    :return: The unquoted value extracted as an unicode
    string that the represents the same value.
    """

    value = legacy.unquote(value, *args, **kwargs)
    is_bytes = type(value) == legacy.BYTES
    if is_bytes: value = value.decode("utf-8")
    return value

def is_content_type(data, target):
    """
    Verifies if the any of the provided mime types (target) is
    valid for the provided content type string.

    :type data: String
    :param data: The content type string to be parsed and matched
    against the target mime type values.
    :type target: Tuple/String
    :param target: The tuple containing the multiple mime type values
    to be verified against the content type mime strings.
    :rtype: bool
    :return: If any of the provided mime types is considered valid
    for the content type.
    """

    if not isinstance(target, (list, tuple)): target = (target,)
    mime, _extra = parse_content_type(data)
    for item in target:
        type, _sub_type = item.split("/")
        wildcard = type + "/*"
        if item in mime: return True
        if wildcard in mime: return True
    return False

def parse_content_type(data):
    """
    Parses the provided content type string retrieving both the multiple
    mime types associated with the resource and the extra key to value
    items associated with the string in case they are defined (it's optional).

    :type data: String
    :param data: The content type data that is going to be parsed to
    obtain the structure of values for the content type string, this must
    be a plain unicode string and not a binary string.
    :rtype: Tuple
    :return: The sequence of mime types of the the content and the multiple
    extra values associated with the content type (eg: charset, boundary, etc.)
    """

    # creates the list of final normalized mime types and the
    # dictionary to store the extra values.
    types = []
    extra_m = dict()

    # in case no valid type has been sent returns the values
    # immediately to avoid further problems
    if not data: return types, extra_m

    # extracts the mime and the extra parts from the data string
    # they are the basis of the processing method
    data = data.strip(";")
    parts = data.split(";")
    mime = parts[0]
    extra = parts[1:]
    mime = mime.strip()

    # runs a series of verifications on the base mime value and in
    # case it's not valid returns the default values immediately
    if not "/" in mime: return types, extra_m

    # strips the complete set of valid extra values, note
    # that these values are going to be processed as key
    # to value items
    extra = [value.strip() for value in extra if extra]

    # splits the complete mime type into its type and sub
    # type components (first step of normalization)
    type, sub_type = mime.split("/", 1)
    sub_types = sub_type.split("+")

    # iterates over the complete set of sub types to
    # create the full mime type for each of them and
    # add the new full items to the types list (normalization)
    for sub_type in sub_types:
        types.append(type + "/" + sub_type)

    # goes through all of the extra key to value items
    # and converts them into proper dictionary values
    for extra_item in extra:
        if not "=" in extra_item: continue
        extra_item = extra_item.strip()
        key, value = extra_item.split("=")
        extra_m[key] = value

    # returns the final tuple containing both the normalized
    # mime types for the content and the extra key to value items
    return types, extra_m

def verify(condition, message = None, code = None, exception = None):
    if condition: return
    exception = exception or exceptions.AssertionError
    kwargs = dict()
    if not code == None: kwargs["code"] = code
    raise exception(
        message or "Assertion of data failed",
        **kwargs
    )

def execute(args, command = None, path = None, shell = True, encoding = None):
    if not encoding: encoding = sys.getfilesystemencoding()
    if command: args = command.split(" ")
    process = subprocess.Popen(
        args,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE,
        shell = shell,
        cwd = path
    )
    code = process.wait()
    stdout = process.stdout.read()
    stderr = process.stderr.read()
    stdout = stdout.decode(encoding)
    stderr = stderr.decode(encoding)
    return dict(
        stdout = stdout,
        stderr = stderr,
        code = code
    )

def _serialize(value):
    if value in legacy.STRINGS: return value
    return json.dumps(value)

class JSONEncoder(json.JSONEncoder):

    def __init__(self, *args, **kwargs):
        self.permissive = kwargs.pop("permissive", True)
        json.JSONEncoder.__init__(self, *args, **kwargs)

    def default(self, obj, **kwargs):
        if hasattr(obj, "json_v"): return obj.json_v()
        if self.permissive: return str(obj)
        return json.JSONEncoder.default(self, obj, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import flask
import functools

from . import exceptions

def errors_json(function):

    @functools.wraps(function)
    def interceptor(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except exceptions.ValidationError as error:
            invalid_m = {}
            for name, _errors in error.errors.items():
                invalid_m[name] = _errors[0]

            return flask.Response(
                json.dumps({
                    "error" : error.message,
                    "invalid" : invalid_m,
                    "exception" : {
                        "message" : error.message,
                        "errors" : error.errors
                    }
                }),
                status = error.code,
                mimetype = "application/json"
            )
        except exceptions.OperationalError as error:
            return flask.Response(
                json.dumps({
                    "error" : error.message,
                    "exception" : {
                        "message" : error.message
                    }
                }),
                status = error.code,
                mimetype = "application/json"
            )

    return interceptor



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import util
from . import exceptions

try: import boto
except: boto = None

connection = None
""" The global wide connection to the amazon server
that is meant to be used across sessions """

bucket = None
""" The global bucket reference to be used to create
new key values and to retrieve existing ones, only one
bucket exists for a given quorum diffusion scope """

id = None
""" The id of the client that is going to be used in the
connections to be established to the amazon servers """

secret = None
""" The secret of the client that is going to be used in the
connections to be established to the amazon servers """

bucket_name = None
""" The name of the bucket to be used to create and retrieve
keys from the amazon services """

def get_connection():
    global connection
    if boto == None: raise exceptions.ModuleNotFound("boto")
    if connection: return connection
    connection = _boto().connect_s3(id, secret)
    return connection

def get_bucket():
    global bucket
    connection = get_connection()
    if bucket: return bucket
    bucket = connection.get_bucket(bucket_name)
    return bucket

def clear_bucket():
    bucket = get_bucket()
    keys = bucket.get_all_keys()
    for key in keys: key.delete()

def get_key(name):
    bucket = get_bucket()
    key = _boto().s3.key.Key(bucket)
    key.key = name
    return key

def exists_key(name):
    bucket = get_bucket()
    key = _boto().s3.key.Key(bucket)
    key.key = name
    return key.exists()

def delete_key(name):
    bucket = get_bucket()
    bucket.delete_key(name)

def _boto(verify = True):
    if verify: util.verify(
        not boto == None,
        message = "boto library not available",
        exception = exceptions.OperationalError
    )
    return boto



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json
import flask
import traceback

from . import log
from . import model
from . import common
from . import legacy
from . import mongodb
from . import exceptions

def route(*args, **kwargs):
    # verifies if the request decorator should be of type
    # json serializer in case it should not returns the old
    # route decorator (default behavior)
    is_json = kwargs.get("json", False)
    if not is_json: return common.base().APP.old_route(*args, **kwargs)

    # removes the json keyword argument from the list of arguments
    # (to avoid errors) and then calls the old route method to obtain
    # the "normal" decorator
    del kwargs["json"]
    decorator = common.base().APP.old_route(*args, **kwargs)

    # creates the "new" route decorator maker method that should
    # override the old one and create a new decorator that
    # serializes all the unhandled exceptions as json
    def _route(function):
        def _decorator(*args, **kwargs):
            try: result = function(*args, **kwargs)
            except exceptions.OperationalError as exception:
                # runs the default handling of the user exception in the flask
                # infra-structure so that the proper exception callbacks are called
                # in case they exist and are properly registered
                common.base().APP.handle_user_exception(exception)

                # prints a warning message to the current logger about the exception
                # that is being handled, note that the traceback is also logger, allowing
                # further debugging information to be printed (extra traceability)
                log.info(
                    "Operational problem while routing request - %s" % legacy.UNICODE(exception),
                    log_trace = True
                )

                # retrieves the formatted version of the traceback information and then
                # splits this information into the various lines that are contained in it
                formatted = traceback.format_exc()
                lines = formatted.splitlines()

                # creates a new response object containing the information about the exception
                # this is going to include the proper exception message, code and also the set
                # of lines that compose the traceback of the exception
                return flask.Response(
                    json.dumps({
                        "exception" : {
                            "name" : exception.__class__.__name__,
                            "message" : exception.message,
                            "code" : exception.code,
                            "traceback" : lines
                        }
                    }),
                    status = exception.code,
                    mimetype = "application/json"
                )
            except BaseException as exception:
                # runs the default handling of the user exception in the flask
                # infra-structure so that the proper exception callbacks are called
                # in case they exist and are properly registered
                common.base().APP.handle_user_exception(exception)

                # prints a warning message to the current logger about the exception
                # that is being handled, note that the traceback is also logger, allowing
                # further debugging information to be printed (extra traceability)
                log.warning(
                    "Problem while routing request - %s" % legacy.UNICODE(exception),
                    log_trace = True
                )

                # retrieves the formatted version of the traceback information and then
                # splits this information into the various lines that are contained in it
                formatted = traceback.format_exc()
                lines = formatted.splitlines()

                # creates the response object that is going to be returned to the client indicating
                # that an undefined exception has just been raised, note that the error code
                # is the internal server error one (undefined exception)
                return flask.Response(
                    json.dumps({
                        "exception" : {
                            "name" : exception.__class__.__name__,
                            "message" : str(exception),
                            "code" : 500,
                            "traceback" : lines
                        }
                    }),
                    status = 500,
                    mimetype = "application/json"
                )

            # retrieves the type for the result that was returned from the
            # concrete method and in case the result is either a mongo object,
            # a dictionary or a sequence it's serialized as json, then returns
            # the result to the caller method
            result_t = type(result)
            if isinstance(result, model.Model):
                result = flask.Response(
                    json.dumps(result),
                    mimetype = "application/json"
                )
            elif mongodb.is_mongo(result):
                result = flask.Response(
                    json.dumps(result),
                    mimetype = "application/json"
                )
            elif result_t in (dict, list, tuple, type(None)):
                result = flask.Response(
                    json.dumps(result),
                    mimetype = "application/json"
                )
            return result

        # verifies if the base value is set in the function for such
        # situations (double decoration) the base value should be set
        # as the already existing base decorator method otherwise keeps
        # using the clojure based one
        has_base = hasattr(function, "_base")
        if has_base: _base = function._base
        else: _base = _decorator

        # updates the decorates with the base value, this is going to be
        # the clojure based new view function, then in case this is a new
        # decorator updates the name of the decorator name with the name
        # of the base function (view function) that is getting decorated
        # so that the original name persists (otherwise a problem would
        # occur in werkzeug)
        _decorator = _base
        if not has_base:
            _decorator.__name__ = function.__name__

        # runs the decorator to create a new decorator and sets the base
        # value in it as the current base value so that it gets propagated
        # all the way to the top decorators in sequence (pile of decorators)
        decorated = decorator(_decorator)
        decorated._base = _base
        return decorated

    return _route



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import json
import time
import flask
import atexit
import logging
import inspect
import datetime
import functools

import werkzeug.debug

from . import acl
from . import log
from . import amqp
from . import util
from . import data
from . import mail
from . import route
from . import model
from . import legacy
from . import amazon
from . import extras
from . import config
from . import session
from . import redisdb
from . import mongodb
from . import pusherc
from . import request
from . import template
from . import execution
from . import exceptions

APP = None
""" The reference to the top level application
that is being handled by quorum """

RUN_CALLED = False
""" Flag that controls if the on run methods have already
been called for the current execution environment """

RUN_F = {}
""" The map that will contain the various functions that
will be called upon the start of the main run loop """

ESCAPE_EXTENSIONS = (
    ".xml",
    ".html",
    ".xhtml",
    ".xml.tpl",
    ".html.tpl",
    ".xhtml.tpl"
)
""" The sequence containing the various extensions
for which the autoescape mode will be enabled  by
default as expected by the end developer """

class Quorum(flask.Flask):
    """
    The top level application class that inherits from the
    flask based one. Should be responsible for the implementation
    of some of the modifications to the original application
    object infra-structure.
    """

    def select_jinja_autoescape(self, filename):
        if filename == None: return False
        if filename.endswith(ESCAPE_EXTENSIONS): return True
        return flask.Flask.select_jinja_autoescape(self, filename)

def monkey():
    """
    Runs the dirty job of monkey patching the flask module
    so that some of its functions get redirected to new
    quorum functions that add extra behavior.

    This is an internal function and should be changed only
    when handling the internals of the quorum framework.
    """

    flask._render_template = flask.render_template
    flask.render_template = template.render_template
    json._default_encoder = util.JSONEncoder()

def call_run():
    global RUN_CALLED
    if RUN_CALLED: return
    for _fname, f in RUN_F.items(): f()
    RUN_CALLED = True

def run(server = None, fallback = "base"):
    if not APP: raise exceptions.BaseError("Application not set or runnable")

    server = config.conf("SERVER", server) or "base"
    runner = globals().get("run_" + server, None)
    runner_f = globals().get("run_" + fallback, None)
    if not runner: raise exceptions.BaseError("Server '%s' not found" % server)

    call_run()

    try: runner()
    except exceptions.ServerInitError as error:
        APP.logger.warning(
            "Server '%s' failed to start (%s) falling back to '%s'" % (
                server, legacy.UNICODE(error), fallback
            )
        )
        runner_f and runner_f()

def prepare_app():
    """
    Prepares the global application object encapsulating
    it with the proper decorators so that it is enabled
    with the proper features (eg: debugging capabilities).

    This method should be called and the returned application
    object used instead of the global one.

    :rtype: Application
    :return: The decorated application object with the proper\
    capabilities enabled, this object should be used for the\
    serving operations instead of the global one.
    """

    app = APP
    if app.debug: app = werkzeug.debug.DebuggedApplication(app, True)
    return app

def run_base():
    debug = config.conf("DEBUG", False, cast = bool)
    reloader = config.conf("RELOADER", False, cast = bool)
    host = config.conf("HOST", "127.0.0.1")
    port = int(config.conf("PORT", 5000))
    APP.run(
        use_debugger = debug,
        debug = debug,
        use_reloader = reloader,
        host = host,
        port = port
    )

def run_netius():
    try: import netius.servers
    except BaseException as exception:
        raise exceptions.ServerInitError(
            legacy.UNICODE(exception),
            server = "netius"
        )

    kwargs = dict()
    host = config.conf("HOST", "127.0.0.1")
    port = int(config.conf("PORT", 5000))
    ipv6 = config.conf("IPV6", False, cast = bool)
    ssl = int(config.conf("SSL", 0)) and True or False
    key_file = config.conf("KEY_FILE", None)
    cer_file = config.conf("CER_FILE", None)
    servers = config.conf_prefix("SERVER_")
    for name, value in servers.items():
        name_s = name.lower()[7:]
        kwargs[name_s] = value
    kwargs["handlers"] = get_handlers()
    kwargs["level"] = get_level()

    # prepares the application object so that it becomes ready
    # to be executed by the server in the proper way
    app = prepare_app()

    # creates the netius wsgi server reference using the provided
    # application object and the constructed keyword arguments
    # and then call the serve method starting the event loop with
    # the proper network configuration
    server = netius.servers.WSGIServer(app, **kwargs)
    server.serve(
        host = host,
        port = port,
        ipv6 = ipv6,
        ssl = ssl,
        key_file = key_file,
        cer_file = cer_file
    )

def run_waitress():
    try: import waitress
    except BaseException as exception:
        raise exceptions.ServerInitError(
            legacy.UNICODE(exception),
            server = "waitress"
        )

    host = config.conf("HOST", "127.0.0.1")
    port = int(config.conf("PORT", 5000))

    # prepares the application object so that it becomes ready
    # to be executed by the server in the proper way
    app = prepare_app()

    # starts the serving process for the waitress server with
    # the proper network configuration values, note that no ssl
    # support is currently available for waitress
    waitress.serve(
        app,
        host = host,
        port = port
    )

def load(
    app = None,
    name = None,
    locales = ("en_us",),
    secret_key = None,
    execution = True,
    redis_session = False,
    mongo_database = None,
    logger = None,
    models = None,
    safe = False,
    **kwargs
):
    """
    Initial loader function responsible for the overriding of
    the flask loading system and for the loading of configuration.

    .. note::

        This function should be called inside you main app file failure
        to do so may result in unexpected behavior.

    :type app: Application
    :param app: The optional flask application object to be used\
    in the loading of quorum (useful for self managed apps).
    :type name: String
    :param name: The name to be used to describe the application\
    for the management of internal values.
    :type locales: List
    :param locales: The list containing the various locale strings for\
    the locales available for the application to be loaded.
    :type secret_key: String
    :param secret_key: The secret seed value to be used for cryptographic\
    operations under flask (eg: client side sessions) this value should\
    be shared among all the pre-fork instances.
    :type execution: bool
    :param execution: Flag indicating if the (background) execution thread\
    should be started providing the required support for background tasks.
    :type redis_session: bool
    :param redis_session: If the session management for the flask infra-structure\
    should be managed using a server side session with support from redis.
    :type mongo_database: String
    :param mongo_database: The default name of the database to be used when\
    using the mongo infra-structure.
    :type logger: String
    :param logger: The name to be used as identification for possible logger\
    files created by the logging sub-module.
    :type models: Module
    :param models: The module containing the complete set of model classes to\
    be used by the data infra-structure (eg: ``mongo``).
    :type safe: bool
    :param safe: If the application should be run in a safe mode meaning that\
    extra validations will be done to ensure proper execution, typically these\
    kind of validations have a performance impacts (not recommended).
    :rtype: Application
    :return: The application that is used by the loaded quorum environment in\
    case one was provided that is retrieved, otherwise the newly created one is\
    returned instead.
    """

    global APP
    if APP: return APP

    # runs the initial loading of the configuration from all
    # the currently available sources (eg: file, environment, etc.)
    load_all()

    # retrieves the value of all the configuration considered
    # to be base and that are going to be used in the loading
    # of the current application (with default values)
    debug = config.conf("DEBUG", False, cast = bool)
    reloader = config.conf("RELOADER", False, cast = bool)
    level_s = config.conf("LEVEL", "WARNING")
    adapter_s = config.conf("ADAPTER", "mongo")
    name = config.conf("NAME", name)
    instance = config.conf("INSTANCE", None)
    force_ssl = config.conf("FORCE_SSL", False)
    redis_url = config.conf("REDISTOGO_URL", None)
    mongo_url = config.conf("MONGOHQ_URL", None)
    amqp_url = config.conf("AMQP_URL", None)
    amazon_id = config.conf("AMAZON_ID", None)
    amazon_secret = config.conf("AMAZON_SECRET", None)
    amazon_bucket = config.conf("AMAZON_BUCKET", None)
    pusher_app_id = config.conf("PUSHER_APP_ID", None)
    pusher_key = config.conf("PUSHER_KEY", None)
    pusher_secret = config.conf("PUSHER_SECRET", None)
    smtp_host = config.conf("SMTP_HOST", None)
    smtp_user = config.conf("SMTP_USER", None)
    smtp_password = config.conf("SMTP_PASSWORD", None)
    redis_url = config.conf("REDIS_URL", redis_url)
    mongo_url = config.conf("MONGOLAB_URI", mongo_url)
    mongo_url = config.conf("MONGO_URL", mongo_url)
    amqp_url = config.conf("CLOUDAMQP_URL", amqp_url)
    amqp_url = config.conf("RABBITMQ_URL", amqp_url)

    # retrieves the possible base URL configuration value and uses it
    # as the basis for the creation of the static URL values to be used
    # by the Flask infra-structure when using _external parameter
    base_url = config.conf("BASE_URL", None)
    if base_url:
        base_parse = legacy.urlparse(base_url)
        is_secure = base_parse.scheme == "https"
        default_port = 443 if is_secure else 80
        server_name = "%s:%d" % (base_parse.hostname, base_parse.port or default_port)
        url_scheme = base_parse.scheme
        application_root = base_parse.path
    else:
        server_name, url_scheme, application_root = None, None, None

    # tries to retrieve the more specific URL related configuration values
    # defaulting to the base URL calculated ones in case they do not exist
    server_name = config.conf("SERVER_NAME", server_name)
    application_root = config.conf("APPLICATION_ROOT", application_root)
    url_scheme = config.conf("URL_SCHEME", url_scheme)
    url_scheme = config.conf("PREFERRED_URL_SCHEME", url_scheme)

    # re-sets the URL related configuration values, making sure that they
    # are properly set in the current configuration environment as they
    # are critical for proper external URL value generation
    config.confs("SERVER_NAME", server_name)
    config.confs("APPLICATION_ROOT", application_root)
    config.confs("PREFERRED_URL_SCHEME", url_scheme)

    # creates the proper values according to the currently provided
    # ones so that they match the ones that are expected
    name = name + "-" + instance if instance else name
    prefix = instance + "-" if instance else ""
    suffix = "-" + instance if instance else ""
    level = logging.DEBUG if debug else _level(level_s)
    logger = logger and prefix + logger

    # retrieves the last stack element as the previous element and
    # uses it to retrieve the module that has triggered the loading
    previous = inspect.stack()[1]
    module = inspect.getmodule(previous[0])

    # uses the module to retrieve the base path for the execution of
    # the app, this is going to be used to calculate relative paths
    path = os.path.dirname(module.__file__)

    # creates the initial app reference using the provided one or
    # creates a new one from the provided/computed name, then sets
    # the current app reference as the current global one
    app = app or Quorum(name)
    APP = app

    # loads the app configuration from the provided keyword arguments
    # map and from the current dictionary of configuration values
    # (value propagation) and then starts the logging process with the
    # requested logger and provided "verbosity" level
    load_app_config(app, kwargs)
    load_app_config(app, config.confd())
    start_log(app, name = logger, level = level)

    # loads the various paths associated with the application into the
    # current environment to reduce the amount of issues related with
    # the importing of modules and other resources
    load_paths(app)

    # loads the complete set of bundle localized in the proper path into
    # the current app environment, this is a blocking operation and may
    # take some time to be performed completely
    load_bundles(app)

    # converts the naming of the adapter into a capital case one and
    # then tries to retrieve the associated class for proper instantiation
    # in case the class is not found the default value is set instead
    adapter_s = adapter_s.capitalize() + "Adapter"
    if not hasattr(data, adapter_s): adapter_s = "MongoAdapter"
    app.adapter = getattr(data, adapter_s)()

    # sets the various eval context filters as such by setting their eval
    # context filter flag to true the jinja infra-structure will handle
    # the rest of the operations so that it's properly used
    util.nl_to_br_jinja.evalcontextfilter = True
    util.nl_to_br_jinja.environmentfilter = True
    util.sp_to_nbsp_jinja.evalcontextfilter = True
    util.sp_to_nbsp_jinja.environmentfilter = True

    # updates the base app reference object with a series of operations
    # that should transform it's base execution and behavior
    app.before_request(before_request)
    app.after_request(after_request)
    app.context_processor(context_processor)
    app.template_filter("locale")(util.to_locale)
    app.template_filter("nl_to_br")(util.nl_to_br_jinja)
    app.template_filter("sp_to_nbsp")(util.sp_to_nbsp_jinja)
    app.request_class = request.Request
    app.locales = locales
    app.safe = safe
    app.debug = debug
    app.use_debugger = debug
    app.use_reloader = reloader
    app.models = models
    app.module = module
    app.path = path
    app.secret_key = secret_key
    app.old_route = app.route
    app.route = route.route
    app.jinja_options = dict(
        finalize = finalize
    )
    app._locale_d = locales[0]

    # sets a series of conditional based attributes in both
    # the associated modules and the base app object (as expected)
    if redis_url: redisdb.url = redis_url
    if mongo_url: mongodb.url = mongo_url
    if amqp_url: amqp.url = amqp_url
    if amazon_id: amazon.id = amazon_id
    if amazon_secret: amazon.secret = amazon_secret
    if amazon_bucket: amazon.bucket_name = amazon_bucket
    if pusher_app_id: pusherc.app_id = pusher_app_id
    if pusher_key: pusherc.key = pusher_key
    if pusher_secret: pusherc.secret = pusher_secret
    if smtp_host: mail.SMTP_HOST = smtp_host
    if smtp_user: mail.SMTP_USER = smtp_user
    if smtp_password: mail.SMTP_PASSWORD = smtp_password
    if execution: start_execution()
    if redis_session: app.session_interface = session.RedisSessionInterface(url = redis_url)
    if mongo_database: mongodb.database = mongo_database + suffix
    if models: setup_models(models)
    if force_ssl: extras.SSLify(app)

    # verifies if the module that has called the method is not
    # of type main and in case it's not calls the runner methods
    # immediately so that the proper initialization is done, then
    # returns the app reference object to the caller method
    if not module.__name__ == "__main__": call_run()
    return app

def unload():
    """
    Unloads the current quorum instance, after this call
    no more access to the quorum facilities is allowed.

    A normal setup of the application would never require
    this function to be called directly.

    .. warning::

        Use this function with care as it may result in unexpected
        behavior from a developer point of view.

    .. note::

        After the call to this method most of the functionally of quorum
        will become unavailable until further call to :func:`quorum.load`.
    """

    global APP
    if not APP: return

    if APP.models: teardown_models(APP.models)

    APP = None

def load_all(path = None):
    load_config(3)
    config.load_file(path = path)
    config.load_env()

def load_config(offset = 1, encoding = "utf-8"):
    element = inspect.stack()[offset]
    module = inspect.getmodule(element[0])
    base_folder = os.path.dirname(module.__file__)
    config.load(path = base_folder, encoding = encoding)

def load_app_config(app, configs):
    for name, value in configs.items():
        app.config[name] = value

def load_paths(app):
    if not app.root_path in sys.path: sys.path.insert(0, app.root_path)

def load_bundles(app, offset = 2):
    # creates the base dictionary that will handle all the loaded
    # bundle information and sets it in the current application
    # object reference so that may be used latter on
    bundles = dict()
    app.bundles = bundles

    # inspects the current stack to obtain the reference to the base
    # application module and then uses it to calculate the base path
    # for the application, from there re-constructs the path to the
    # bundle file and verifies its own existence
    element = inspect.stack()[offset]
    module = inspect.getmodule(element[0])
    base_folder = os.path.dirname(module.__file__)
    bundles_path = os.path.join(base_folder, "bundles")
    if not os.path.exists(bundles_path): return

    # list the bundles directory files and iterates over each of the
    # files to load its own contents into the bundles "registry"
    paths = os.listdir(bundles_path)
    for path in paths:
        # joins the current (base) bundles path with the current path
        # in iteration to create the full path to the file and opens
        # it trying to read its json based contents
        path_f = os.path.join(bundles_path, path)
        file = open(path_f, "rb")
        try: data_j = json.load(file)
        except: continue
        finally: file.close()

        # unpacks the current path in iteration into the base name,
        # locale string and file extension to be used in the registration
        # of the data in the bundles registry
        try: _base, locale, _extension = path.split(".", 2)
        except: continue

        # retrieves a possible existing map for the current locale in the
        # registry and updates such map with the loaded data, then re-updates
        # the reference to the locale in the current bundle registry
        bundle = bundles.get(locale, {})
        bundle.update(data_j)
        bundles[locale] = bundle

def start_log(
    app,
    name = None,
    level = logging.WARN,
    format_base = log.LOGGING_FORMAT,
    format_tid = log.LOGGING_FORMAT_TID
):
    # tries to retrieve some of the default configuration values
    # that are going to be used in the logger startup
    format = config.conf("LOGGING_FORMAT", None)

    # "resolves" the proper logger file path taking into account
    # the currently defined operative system, should uses the system
    # level path in case the operative system is unix based
    if os.name == "nt": path_t = "%s"
    else: path_t = "/var/log/%s"
    path = name and path_t % name

    # creates the map that is going to be used to store the
    # various handlers registered for the logger
    app.handlers = dict()

    # creates the formatter object from the provided string
    # so that it may be used in the various handlers
    formatter = log.ThreadFormatter(format or format_base)
    formatter.set_tid(format or format_tid)

    # retrieves the reference to the logger object currently
    # associated with the app and disable the parent in it,
    # then removes the complete set of associated handlers and
    # sets the proper debug level, note that the logger is
    # going to be shared between quorum and flask (common logger)
    logger = app.logger if hasattr(app, "logger") else logging.getLogger("quorum")
    logger.parent = None
    logger.handlers = []
    logger.setLevel(level)

    # creates both the stream and the memory based handlers that
    # are going to be used for the current logger
    stream_handler = logging.StreamHandler()
    memory_handler = log.MemoryHandler()

    try:
        # tries to create the file handler for the logger with the
        # resolve path (operation may fail for permissions)
        file_handler = path and logging.FileHandler(path)
    except:
        # in case there's an error creating the file handler for
        # the logger prints an error message indicating the problem
        sys.stderr.write("Problem starting logging for file '%s'\n" % path)
        file_handler = None

    # adds the various created handler to the current logger so that
    # they are going to be used when using the logger for output
    if stream_handler: logger.addHandler(stream_handler)
    if memory_handler: logger.addHandler(memory_handler)
    if file_handler: logger.addHandler(file_handler)

    # for each of the handlers adds them to the handlers map in case
    # they are valid and defined (no problem in construction)
    if stream_handler: app.handlers["stream"] = stream_handler
    if memory_handler: app.handlers["memory"] = memory_handler
    if file_handler: app.handlers["file"] = file_handler

    # iterates over the complete set of handlers currently
    # registered in the logger to properly set the formatter
    # and the level for all of them (as specified)
    for handler in logger.handlers:
        handler.setFormatter(formatter)
        handler.setLevel(level)

    # runs the extra logging step for the current state, meaning that
    # some more handlers may be created according to the logging config
    extra_logging(logger, level, formatter)

    # sets the current logger in the top level app value so that
    # this logger is going to be used as the quorum logger
    app.logger_q = logger

def extra_logging(logger, level, formatter):
    """
    Loads the complete set of logging handlers defined in the
    current logging value, should be a map of definitions.

    This handlers will latter be used for piping the various
    logging messages to certain output channels.

    The creation of the handler is done using a special keyword
    arguments strategy so that python and configuration files
    are properly set as compatible.

    :type logger Logger
    :param logger: The logger currently in use for where the new\
    handlers that are going to be created will be added.
    :type level: String/int
    :param level: The base severity level for which the new handler\
    will be configured in case no extra level definition is set.
    :type formatter: Formatter
    :param formatter: The logging formatter instance to be set in\
    the handler for formatting messages to the output.
    """

    # verifies if the logging attribute of the current instance is
    # defined and in case it's not returns immediately
    logging = config.conf("LOGGING", None)
    if not logging: return

    # iterates over the complete set of handler configuration in the
    # logging to create the associated handler instances
    for _config in logging:
        # gathers the base information on the current handler configuration
        # running also the appropriate transformation on the level
        name = _config.get("name", None)
        __level = _config.get("level", level)
        __level = _level(__level)

        # "clones" the configuration dictionary and then removes the base
        # values so that they do not interfere with the building
        _config = dict(_config)
        if "level" in _config: del _config["level"]
        if "name" in _config: del _config["name"]

        # retrieves the proper building, skipping the current loop in case
        # it does not exits and then builds the new handler instance, setting
        # the proper level and formatter and then adding it to the logger
        if not hasattr(log, name + "_handler"): continue
        builder = getattr(log, name + "_handler")
        handler = builder(**_config)
        handler.setLevel(__level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

def get_app(app = None):
    return app or APP

def get_adapter(app = None):
    return APP and APP.adapter

def get_log(app = None):
    app = app or APP
    if not app: return None
    is_custom = hasattr(app, "logger_q")
    return app.logger_q if is_custom else app.logger

def get_level(app = None):
    logger = get_log(app = app)
    if not logger: return None
    return logger.level

def get_handlers(app = None):
    logger = get_log(app = app)
    if not logger: return None
    return logger.handlers

def get_handler(name, app = None):
    app = app or APP
    if not app: return None
    return app.handlers.get(name, None)

def get_bundle(name, app = None):
    app = app or APP
    if not app: return None
    bundle = app.bundles.get(name, None)
    if bundle: return bundle
    name = _best_locale(name)
    return app.bundles.get(name, None)

def is_devel(app = None):
    level = get_level(app = app)
    if not level: return False

def finalize(value):
    # returns an empty string as value representation
    # for unset values, this is the default representation
    # to be used in the template engine
    if value == None: return ""
    return value

def before_request():
    flask.request.args_s = util.load_form(flask.request.args)
    flask.request.form_s = util.load_form(flask.request.form)
    flask.request.locale = util.load_locale(APP.locales)
    util.set_locale()

def after_request(response):
    if APP.safe: util.reset_locale()
    util.anotate_async(response)
    return response

def context_processor():
    return dict(
        acl = acl.check_login,
        conf = config.conf,
        locale = util.to_locale,
        nl_to_br = util.nl_to_br,
        sp_to_nbsp = util.sp_to_nbsp,
        date_time = util.date_time,
        time = time,
        datetime = datetime,
        zip = zip
    )

def start_execution():
    # creates the thread that it's going to be used to
    # execute the various background tasks and starts
    # it, providing the mechanism for execution
    execution.background_t = execution.ExecutionThread()
    background_t = execution.background_t
    background_t.start()

@atexit.register
def stop_execution():
    # stop the execution thread so that it's possible to
    # the process to return the calling
    background_t = execution.background_t
    background_t and background_t.stop()

def setup_models(models):
    _models_c = models_c(models = models)
    for model_c in _models_c: model_c.setup()

def teardown_models(models):
    _models_c = models_c(models = models)
    for model_c in _models_c: model_c.teardown()

def models_c(models = None):
    # retrieves the proper models defaulting to the current
    # application models in case they are not defined, note
    # that in case the model is not defined an empty list is
    # going to be returned (fallback process)
    models = models or APP.models
    if not models: return []

    # creates the list that will hold the various model
    # class discovered through module analysis
    models_c = []

    # iterates over the complete set of items in the models
    # modules to find the ones that inherit from the base
    # model class for those are the real models
    for _name, value in models.__dict__.items():
        # verifies if the current value in iteration inherits
        # from the top level model in case it does not continues
        # the loop as there's nothing to be done
        try: is_valid = issubclass(value, model.Model)
        except: is_valid = False
        if not is_valid: continue

        # adds the current value in iteration as a new class
        # to the list that hold the various model classes
        models_c.append(value)

    # returns the list containing the various model classes
    # to the caller method as expected by definition
    return models_c

def resolve(identifier = "_id", counters = True):
    # creates the list that will hold the definition of the current
    # model classes with a sequence of name and identifier values
    entities = []

    # retrieves the complete set of model classes registered
    # for the current application and for each of them retrieves
    # the name of it and creates a tuple with the name and the
    # identifier attribute name adding then the tuple to the
    # list of entities tuples (resolution list)
    _models_c = models_c()
    for model_c in _models_c:
        name = model_c._name()
        tuple = (name, identifier)
        entities.append(tuple)

    # in case the counters flag is defined the counters tuple containing
    # the counters table name and identifier is added to the entities list
    if counters: entities.append(("counters", identifier))

    # returns the resolution list to the caller method as requested
    # by the call to this method
    return entities

def templates_path():
    return os.path.join(APP.root_path, APP.template_folder)

def bundles_path():
    return os.path.join(APP.root_path, "bundles")

def base_path(*args, **kwargs):
    return os.path.join(APP.root_path, *args)

def has_context():
    return True if flask._app_ctx_stack.top else False

def ensure_context(function):
    """
    Decorator that makes sure that the underlying execution
    method/function is run inside a valid app context.

    In case there's currently no app context defined it uses
    the global Application reference to create a new one.

    :type function: Function
    :param function: The function that is going to be decorated
    by the current ensure decorator.
    :rtype: Decorator
    :return: The decorator that should be used for the ensuring
    of the app context in the current execution environment.
    """

    @functools.wraps(function)
    def interceptor(*args, **kwargs):
        _ctx = has_context()
        _app_ctx = APP.app_context() if APP else None
        _ensure = True if not _ctx and _app_ctx else False
        try:
            if _ensure: flask._app_ctx_stack.push(_app_ctx)
            result = function(*args, **kwargs)
        finally:
            if _ensure: flask._app_ctx_stack.pop()
        return result

    return interceptor

def onrun(function):
    fname = function.__name__
    if fname in RUN_F: return
    RUN_F[fname] = function
    return function

def _level(level):
    """
    Converts the provided logging level value into the best
    representation of it, so that it may be used to update
    a logger's level of representation.

    This method takes into account the current interpreter
    version so that no problem occur.

    :type level: String/int
    :param level: The level value that is meant to be converted
    into the best representation possible.
    :rtype: int
    :return: The best representation of the level so that it may
    be used freely for the setting of logging levels under the
    current running interpreter.
    """

    level_t = type(level)
    if level_t == int: return level
    if level == None: return level
    if level == "SILENT": return log.SILENT
    if hasattr(logging, "_checkLevel"):
        return logging._checkLevel(level)
    return logging.getLevelName(level)

def _best_locale(locale, app = None):
    if not locale: return locale
    app = app or APP
    for _locale in app.locales:
        is_valid = _locale.startswith(locale)
        if not is_valid: continue
        return _locale
    return locale

# runs the monkey patching of the flask module so that it
# may be used according to the quorum specification, this
# is used in order to avoid minimal effort in the conversion
# of old flask based applications (reverse compatibility)
monkey()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import legacy

class Ordered(type):
    """
    Metaclass to be used for classes where the
    definition order of its attributes is critical
    for the correct process.
    """

    def __new__(cls, name, bases, attrs):
        new_cls = super(Ordered, cls).__new__(cls, name, bases, attrs)
        new_cls._ordered = [(name, attrs.pop(name)) for name, value in\
            legacy.eager(attrs.items()) if hasattr(value, "creation_counter")]
        new_cls._ordered.sort(key = lambda item: item[1].creation_counter)
        new_cls._ordered = [name for name, value in new_cls._ordered]
        return new_cls

    def __init__(cls, name, bases, attrs):
        super(Ordered, cls).__init__(name, bases, attrs)

    def __cmp__(self, value):
        return cmp(self.__name__, value.__name__) #@UndefinedVariable

    def __lt__(self, value):
        return self.__name__.__lt__(value.__name__)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import time
import struct
import socket
import hashlib
import binascii
import threading

from . import log
from . import legacy
from . import config
from . import mongodb
from . import exceptions

class DataAdapter(object):

    def __init__(self, *args, **kwargs):
        self._inc = 0
        self._machine_bytes = self.__machine_bytes()
        self._inc_lock = threading.RLock()

    @classmethod
    def name_g(cls):
        return cls.__name__[:-7].lower()

    def encoder(self):
        return None

    def collection(self, name, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def reset(self):
        raise exceptions.NotImplementedError()

    def get_db(self):
        raise exceptions.NotImplementedError()

    def drop_db(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def object_id(self, value = None):
        if not value: return self._id()
        if not len(value) == 24:
            raise exceptions.OperationalError(
                message = "Expected object id of length 24 chars"
            )
        return value

    @property
    def name(self):
        cls = self.__class__
        return cls.name_g()

    def _id(self):
        token = struct.pack(">i", int(time.time()))
        token += self._machine_bytes
        token += struct.pack(">H", os.getpid() % 0xffff)
        self._inc_lock.acquire()
        try:
            token += struct.pack(">i", self._inc)[1:4]
            self._inc = (self._inc + 1) % 0xffffff
        except:
            self._inc_lock.release()
        token_s = binascii.hexlify(token)
        token_s = legacy.str(token_s)
        return token_s

    def __machine_bytes(self):
        machine_hash = hashlib.md5()
        hostname = socket.gethostname()
        hostname = legacy.bytes(hostname)
        machine_hash.update(hostname)
        return machine_hash.digest()[0:3]

class MongoAdapter(DataAdapter):

    def encoder(self):
        return mongodb.MongoEncoder

    def collection(self, name, *args, **kwargs):
        db = self.get_db()
        collection = db[name]
        return MongoCollection(self, name, collection)

    def reset(self):
        return mongodb.reset_connection()

    def get_db(self):
        return mongodb.get_db()

    def drop_db(self, *args, **kwargs):
        return mongodb.drop_db()

    def object_id(self, value = None):
        if not value: return self._id()
        return mongodb.object_id(value)

    def _id(self):
        return mongodb.object_id(None)

class TinyAdapter(DataAdapter):

    def __init__(self, *args, **kwargs):
        DataAdapter.__init__(self, *args, **kwargs)
        self.file_path = config.conf("TINY_PATH", "db.json")
        self.storage = config.conf("TINY_STORAGE", "json")
        self.file_path = kwargs.get("file_path", self.file_path)
        self._db = None

    def collection(self, name, *args, **kwargs):
        db = self.get_db()
        table = db.table(name)
        return TinyCollection(self, name, table)

    def reset(self):
        pass

    def get_db(self):
        if not self._db == None: return self._db
        method = getattr(self, "_get_db_%s" % self.storage)
        self._db = method()
        return self._db

    def drop_db(self, *args, **kwargs):
        if self._db == None: return
        db = self.get_db()
        db.purge_tables()
        db.close()
        self._db = None
        method = getattr(self, "_drop_db_%s" % self.storage)
        method()

    def _get_db_json(self):
        import tinydb
        return tinydb.TinyDB(self.file_path)

    def _get_db_memory(self):
        import tinydb
        return tinydb.TinyDB(
            storage = tinydb.storages.MemoryStorage
        )

    def _drop_db_json(self):
        os.remove(self.file_path)

    def _drop_db_memory(self):
        pass

class Collection(object):

    def __init__(self, owner, name):
        self.owner = owner
        self.name = name

    def find(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def find_one(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def find_and_modify(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def insert(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def update(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def remove(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def count(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def ensure_index(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def drop_indexes(self, *args, **kwargs):
        raise exceptions.NotImplementedError()

    def object_id(self, *args, **kwargs):
        return self.owner.object_id(*args, **kwargs)

    def log(self, operation, *args, **kwargs):
        show_queries = config.conf("SHOW_QUERIES", False, cast = bool)
        if not show_queries: return
        extra = kwargs or args
        log.debug(
            "[%s] %10s -> %12s <-> %s" %\
            (self.owner.name, operation, self.name, str(extra)[:2046])
        )

    def _id(self, *args, **kwargs):
        return self.owner._id(*args, **kwargs)

class MongoCollection(Collection):

    def __init__(self, owner, name, base):
        Collection.__init__(self, owner, name)
        self._base = base

    def find(self, *args, **kwargs):
        self.log("find", *args, **kwargs)
        return self._base.find(*args, **kwargs)

    def find_one(self, *args, **kwargs):
        self.log("find_one", *args, **kwargs)
        return self._base.find_one(*args, **kwargs)

    def find_and_modify(self, *args, **kwargs):
        self.log("find_and_modify", *args, **kwargs)
        return mongodb._store_find_and_modify(self._base, *args, **kwargs)

    def insert(self, *args, **kwargs):
        self.log("insert", *args, **kwargs)
        return mongodb._store_insert(self._base, *args, **kwargs)

    def update(self, *args, **kwargs):
        self.log("update", *args, **kwargs)
        return mongodb._store_update(self._base, *args, **kwargs)

    def remove(self, *args, **kwargs):
        self.log("remove", *args, **kwargs)
        return mongodb._store_remove(self._base, *args, **kwargs)

    def count(self, *args, **kwargs):
        self.log("count", *args, **kwargs)
        return self._base.count(*args, **kwargs)

    def ensure_index(self, *args, **kwargs):
        self.log("ensure_index", *args, **kwargs)
        direction = kwargs.pop("direction", True)

        is_simple = direction == "simple"
        is_all = direction == "all"
        is_multiple = isinstance(direction, (list, tuple))
        is_direction = not is_all and not is_simple and\
            (legacy.is_string(direction) or type(direction) == int)
        is_single = is_simple or is_direction

        if is_all: kwargs["directions"] = "all"
        if is_multiple: kwargs["directions"] = direction
        if is_direction: args = list(args); args[0] = [(args[0], direction)]

        if is_single: return mongodb._store_ensure_index(self._base, *args, **kwargs)
        else: return mongodb._store_ensure_index_many(self._base, *args, **kwargs)

    def drop_indexes(self, *args, **kwargs):
        self.log("drop_indexes", *args, **kwargs)
        return self._base.drop_indexes()

class TinyCollection(Collection):

    def __init__(self, owner, name, base):
        Collection.__init__(self, owner, name)
        self._base = base

    def find(self, *args, **kwargs):
        self.log("find", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        condition = self._to_condition(filter)
        results = self._base.search(condition)
        return self._to_results(results, kwargs)

    def find_one(self, *args, **kwargs):
        self.log("find_one", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        condition = self._to_condition(filter)
        results = self._base.search(condition)
        results = self._to_results(results, kwargs)
        return results[0] if results else None

    def find_and_modify(self, *args, **kwargs):
        self.log("find_and_modify", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        modification = args[1] if len(args) > 1 else dict()
        create = kwargs.get("new", False)
        condition = self._to_condition(filter)
        object = self._base.get(condition)
        found = True if object else False
        if not found and not create:
            raise exceptions.OperationalError("No object found")
        if not found: object = dict(filter)
        object = self._to_update(modification, object = object)
        if found: self.update(filter, {"$set" : object})
        else: self.insert(object)
        return dict(object)

    def insert(self, *args, **kwargs):
        self.log("insert", *args, **kwargs)
        object = args[0] if len(args) > 0 else dict()
        has_id = "_id" in object
        if not has_id: object["_id"] = self._id()
        self._base.insert(object)
        return object

    def update(self, *args, **kwargs):
        self.log("update", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        updater = args[1] if len(args) > 1 else dict()
        condition = self._to_condition(filter)
        object = updater.get("$set", dict())
        return self._base.update(object, condition)

    def remove(self, *args, **kwargs):
        self.log("remove", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        condition = self._to_condition(filter)
        return self._base.remove(condition)

    def count(self, *args, **kwargs):
        self.log("count", *args, **kwargs)
        filter = args[0] if len(args) > 0 else dict()
        condition = self._to_condition(filter)
        return self._base.count(condition)

    def ensure_index(self, *args, **kwargs):
        self.log("ensure_index", *args, **kwargs)

    def drop_indexes(self, *args, **kwargs):
        self.log("drop_indexes", *args, **kwargs)

    def _to_condition(self, filter):
        import tinydb
        query = tinydb.Query()
        condition = query._id.exists()
        for name, value in legacy.iteritems(filter):
            if name.startswith("$"): continue
            query = tinydb.Query()
            _condition = getattr(query, name).__eq__(value)
            condition &= _condition
        return condition

    def _to_results(self, results, kwargs, build = True):
        sort = kwargs.get("sort", [])
        skip = kwargs.get("skip", 0)
        limit = kwargs.get("limit", None)

        limit = None if limit == 0 else limit
        reverse = sort[0][1] == -1 if sort else False

        def sorter(value):
            result = []
            for item in sort: result.append(value[item[0]])
            return tuple(result)

        if sort: results.sort(key = sorter, reverse = reverse)
        if skip or limit: results = results[slice(skip, skip + limit, 1)]
        if build: results = [dict(result) for result in results]
        return results

    def _to_update(self, modification, object = None):
        object = object or dict()
        increments = modification.get("$inc", {})
        for name, increment in legacy.iteritems(increments):
            value = object.get(name, 0)
            value += increment
            object[name] = value
        return object



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Flask Quorum
# Copyright (c) 2008-2017 Hive Solutions Lda.
#
# This file is part of Hive Flask Quorum.
#
# Hive Flask Quorum is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Flask Quorum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Flask Quorum. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import json
import shelve

from . import util
from . import config
from . import exceptions

try: import redis
except: redis = None

connection = None
""" The global connection object that should persist
the connection relation with the database service """

url = None
""" The global variable containing the URL to be used
for the connection with the service """

class RedisMemory(object):
    """
    "Local" in memory stub object that simulates
    the redis interface, useful for debugging.

    This memory interface may create problems in
    a multiple process environment (non shared memory).
    """

    values = None
    """ The map containing the various values to
    be set in the memory map, simulates the redis
    data store """

    def __init__(self):
        self.values = {}

    def get(self, name):
        name_s = str(name)
        return self.values.get(name_s)

    def set(self, name, value):
        name_s = str(name)
        self.values[name_s] = value

    def setex(self, name, value, expire):
        self.set(name, value)

    def delete(self, name):
        if not name in self.values: return
        del self.values[name]

class RedisShelve(RedisMemory):
    """
    "Local" in persistent stub object that simulates
    the redis interface, useful for debugging.

    This shelve interface requires a writable path
    where its persistent file may be written.
    """

    def __init__(self, path = "redis.shelve"):
        RedisMemory.__init__(self)
        self.db_path = path
        self.open_db()

    def close(self):
        self.values.close()

    def set(self, name, value, secure = None):
        RedisMemory.set(self, name, value)
        if secure == None:
            secure = self.db_secure()
        if secure:
            self.values.close()
            self.open_db()
        else:
            self.values.sync()

    def delete(self, name):
        name_s = str(name)
        if not name_s in self.values: return
        del self.values[name_s]

    def open_db(self):
        base_path = config.conf("SESSION_FILE_PATH", "")
        file_path = os.path.join(base_path, self.db_path)
        self.values = shelve.open(
            file_path,
            protocol = 2,
            writeback = True
        )

    def db_type(self):
        shelve_cls = type(self.values.dict)
        shelve_dbm = shelve_cls.__name__
        return shelve_dbm

    def db_secure(self):
        return self.db_type() == "dbm"

def get_connection():
    return _get_connection(url)

def dumps(*args):
    return json.dumps(*args)

def _get_connection(url):
    global connection
    if redis == None: raise exceptions.ModuleNotFound("redis")
    if not connection: connection = url and _redis().from_url(url) or RedisShelve()
    return connection

def _redis(verify = True):
    if verify: util.verify(
        not redis == None,
        message = "redis library not available",
        exception = exceptions.OperationalError
    )
    return redis
